# -*- coding: utf-8; -*-
"""add report_output.params

Revision ID: abd43933c6c6
Revises: 90d5e8185d44
Create Date: 2019-05-07 16:50:24.936129

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'abd43933c6c6'
down_revision = u'90d5e8185d44'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # report_output
    op.add_column('report_output', sa.Column('report_type', sa.String(length=255), nullable=True))
    op.add_column('report_output', sa.Column('params', rattail.db.types.JSONTextDict(), nullable=True))


def downgrade():

    # report_output
    op.drop_column('report_output', 'params')
    op.drop_column('report_output', 'report_type')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Board Handler
"""

from __future__ import unicode_literals, absolute_import

from rattail.util import load_object
from rattail.app import GenericHandler


class BoardHandler(GenericHandler):
    """
    Base class and default implementation for board handlers.
    """

    def method_not_implemented(self, method):
        msg = "Please implement {}.{}()".format(
            self.__class__.__name__, method)
        return NotImplementedError(msg)

    def begin_board_membership(self, person, start_date, **kwargs):
        """
        Begin board membership for the given person.
        """
        raise self.method_not_implemented('begin_board_membership')

    def end_board_membership(self, person, end_date, **kwargs):
        """
        End board membership for the given person.
        """
        raise self.method_not_implemented('end_board_membership')

    def why_not_begin_board_membership(self, person):
        """
        Inspect the given person and if they should not be made a current board
        member for any reason, return that reason as text.  If it's okay for
        the person to be made a board member, returns ``None``.
        """
        raise self.method_not_implemented('why_not_begin_board_membership')

    def why_not_end_board_membership(self, person):
        """
        Inspect the given person and if their current board membership should
        not be ended for any reason, return that reason as text.  If it's okay
        for the person to stop being a board member, returns ``None``.
        """
        raise self.method_not_implemented('why_not_end_board_membership')


def get_board_handler(config, **kwargs):
    """
    Create and return the configured :class:`BoardHandler` instance.
    """
    spec = config.get('rattail', 'board.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = BoardHandler
    return factory(config, **kwargs)



# -*- coding: utf-8 -*-
"""add batch.notes

Revision ID: 9305c1f493b4
Revises: 528c049eb5b7
Create Date: 2017-02-21 14:07:52.838097

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '9305c1f493b4'
down_revision = u'528c049eb5b7'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # add 'notes' for batch headers
    op.add_column('batch_handheld', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('batch_inventory', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('batch_pricing', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('label_batch', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('purchase_batch', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('vendor_catalog', sa.Column('notes', sa.Text(), nullable=True))
    op.add_column('vendor_invoice', sa.Column('notes', sa.Text(), nullable=True))


def downgrade():

    # drop 'notes' for batch headers
    op.drop_column('vendor_invoice', 'notes')
    op.drop_column('vendor_catalog', 'notes')
    op.drop_column('purchase_batch', 'notes')
    op.drop_column('label_batch', 'notes')
    op.drop_column('batch_pricing', 'notes')
    op.drop_column('batch_inventory', 'notes')
    op.drop_column('batch_handheld', 'notes')



# -*- coding: utf-8; -*-
"""add product.item_id, item_type

Revision ID: 8e965902bf93
Revises: 0a5a866f2a3d
Create Date: 2017-03-23 18:05:00.431253

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '8e965902bf93'
down_revision = u'0a5a866f2a3d'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product
    op.add_column('product', sa.Column('item_id', sa.String(length=20), nullable=True))
    op.add_column('product', sa.Column('item_type', sa.Integer(), nullable=True))


def downgrade():

    # product
    op.drop_column('product', 'item_type')
    op.drop_column('product', 'item_id')



# -*- coding: utf-8 -*-
"""add purchase_item.out_of_stock

Revision ID: adcb91f473c3
Revises: 0a3dca1cd299
Create Date: 2018-09-25 18:53:21.503034

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'adcb91f473c3'
down_revision = u'0a3dca1cd299'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch_row
    op.add_column('purchase_batch_row', sa.Column('out_of_stock', sa.Boolean(), nullable=True))

    # purchase_item
    op.add_column('purchase_item', sa.Column('out_of_stock', sa.Boolean(), nullable=True))

    # vendor_invoice_row
    op.add_column('vendor_invoice_row', sa.Column('out_of_stock', sa.Boolean(), nullable=True))


def downgrade():

    # vendor_invoice_row
    op.drop_column('vendor_invoice_row', 'out_of_stock')

    # purchase_item
    op.drop_column('purchase_item', 'out_of_stock')

    # purchase_batch_row
    op.drop_column('purchase_batch_row', 'out_of_stock')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model for inventory batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.ext.orderinglist import ordering_list

from rattail.db.model import Base, BatchMixin, ProductBatchRowMixin, uuid_column, getset_factory


class InventoryBatch(BatchMixin, Base):
    """
    Batch for product inventory counts; note that this requires a data file.
    """
    __tablename__ = 'batch_inventory'
    __batchrow_class__ = 'InventoryBatchRow'
    batch_key = 'inventory'

    @declared_attr
    def __table_args__(cls):
        return cls.__default_table_args__() + (
            sa.ForeignKeyConstraint(['handheld_batch_uuid'], ['batch_handheld.uuid'],
                                    name='batch_inventory_fk_handheld_batch'),
            )

    handheld_batch_uuid = sa.Column(sa.String(length=32), nullable=True)

    handheld_batch = orm.relationship(
        'HandheldBatch',
        doc="""
        Reference to the handheld batch from which this inventory batch originated.
        """,
        backref=orm.backref('inventory_batch', uselist=False, doc="""
        Reference to the inventory batch to which this handheld batch was converted.
        """))

    mode = sa.Column(sa.Integer(), nullable=True, doc="""
    Specifies the "mode" for the inventory count batch, i.e. how the count data
    should ultimately be interpreted/applied.
    """)

    reason_code = sa.Column(sa.String(length=50), nullable=True, doc="""
    Specifies the "reason code" for an inventory adjustment batch, i.e.  what
    happened / why the adjustment is required.  Interpretation of this value is
    the responsibility of custom logic.
    """)

    total_cost = sa.Column(sa.Numeric(precision=12, scale=5), nullable=True, doc="""
    Total inventory cost for the batch, if known.
    """)


class InventoryBatchFromHandheld(Base):
    """
    Primary data model for inventory batches.
    """
    __tablename__ = 'batch_inventory_handheld'
    __table_args__ = (
        sa.ForeignKeyConstraint(['batch_uuid'], ['batch_inventory.uuid'],
                                name='batch_inventory_handheld_fk_batch'),
        sa.ForeignKeyConstraint(['handheld_uuid'], ['batch_handheld.uuid'],
                                name='batch_inventory_handheld_fk_handheld'),
    )

    uuid = uuid_column()

    batch_uuid = sa.Column(sa.String(length=32), nullable=False)
    ordinal = sa.Column(sa.Integer(), nullable=False)
    batch = orm.relationship(
        InventoryBatch,
        doc="""
        Reference to the inventory batch, with which handheld batch(es) are associated.
        """,
        backref=orm.backref(
            '_handhelds',
            collection_class=ordering_list('ordinal', count_from=1),
            order_by=ordinal,
            cascade='all, delete-orphan',
            doc="""
            Sequence of raw inventory / handheld batch associations.
            """))
    
    handheld_uuid = sa.Column(sa.String(length=32), nullable=False)
    handheld = orm.relationship(
        'HandheldBatch',
        doc="""
        Reference to the handheld batch from which this inventory batch originated.
        """,
        backref=orm.backref(
            '_inventory_batch',
            uselist=False,
            cascade='all, delete-orphan',
            doc="""
            Indirect reference to the inventory batch to which this handheld batch was converted.
            """))


InventoryBatch.handheld_batches = association_proxy('_handhelds', 'handheld',
                                                    creator=lambda batch: InventoryBatchFromHandheld(handheld=batch),
                                                    getset_factory=getset_factory)


class InventoryBatchRow(ProductBatchRowMixin, Base):
    """
    Rows for inventory batches.
    """
    __tablename__ = 'batch_inventory_row'
    __batch_class__ = InventoryBatch

    STATUS_OK = 1
    STATUS_PRODUCT_NOT_FOUND = 2

    STATUS = {
        STATUS_OK:                      "ok",
        STATUS_PRODUCT_NOT_FOUND:       "product not found",
    }

    previous_units_on_hand = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Previous on-hand unit quantity for the item, i.e. before the count.
    """)

    cases = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Case quantity for the record.
    """)

    units = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Unit quantity for the record.
    """)

    case_quantity = sa.Column(sa.Numeric(precision=6, scale=2), nullable=True, doc="""
    Number of units in a case of product.
    """)

    variance = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    This is essentially the difference between the "supposed" number of units
    on-hand at the time of count, and the actual number of units counted.  May
    be useful for reporting, or for a "delayed" adjustment to actual inventory
    levels.
    """)

    unit_cost = sa.Column(sa.Numeric(precision=9, scale=5), nullable=True, doc="""
    Inventory unit cost for the item, if known.
    """)

    total_cost = sa.Column(sa.Numeric(precision=12, scale=5), nullable=True, doc="""
    Total inventory cost for the item (i.e. taking quantity into account), if known.
    """)

    @property
    def full_unit_quantity(self):
        if self.cases is not None or self.units is not None:
            return self.units or 0 + (self.cases or 0) * (self.case_quantity or 1)

    @property
    def case_cost(self):
        if self.case_quantity and self.unit_cost:
            return self.case_quantity * self.unit_cost



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Handler for "customer order" batches
"""

from __future__ import unicode_literals, absolute_import, division

import six
from sqlalchemy import orm

from rattail.db import model
from rattail.batch import BatchHandler


class CustomerOrderBatchHandler(BatchHandler):
    """
    Handler for all "customer order" batches, regardless of "mode".  The
    handler must inspect the
    :attr:`~rattail.db.model.batch.custorder.CustomerOrderBatch.mode` attribute
    of each batch it deals with, in order to determine which logic to apply.
    """
    batch_model_class = model.CustomerOrderBatch

    def get_case_size_for_product(self, product):
        return product.case_size

    def refresh_row(self, row):
        if not row.product:
            if row.item_entry:
                session = orm.object_session(row)
                # TODO: should do more than just query for uuid here
                product = session.query(model.Product).get(row.item_entry)
                if product:
                    row.product = product
            if not row.product:
                row.status_code = row.STATUS_PRODUCT_NOT_FOUND
                return

        product = row.product
        row.product_upc = product.upc
        row.product_brand = six.text_type(product.brand or "")
        row.product_description = product.description
        row.product_size = product.size
        row.product_weighed = product.weighed
        row.case_quantity = self.get_case_size_for_product(product)

        department = product.department
        row.department_number = department.number if department else None
        row.department_name = department.name if department else None

        cost = product.cost
        row.product_unit_cost = cost.unit_cost if cost else None

        regprice = product.regular_price
        row.unit_price = regprice.price if regprice else None

        # we need to know if total price is updated
        old_total = row.total_price

        # maybe update total price
        if row.unit_price is None:
            row.total_price = None
        elif not row.unit_price:
            row.total_price = 0
        else:
            row.total_price = row.unit_price * row.order_quantity
            if row.order_uom == self.enum.UNIT_OF_MEASURE_CASE:
                row.total_price *= (row.case_quantity or 1)

        # update total price for batch too, if it changed
        if row.total_price != old_total:
            batch = row.batch
            batch.total_price = ((batch.total_price or 0)
                                 + (row.total_price or 0)
                                 - (old_total or 0))

        row.status_code = row.STATUS_OK

    def remove_row(self, row):
        batch = row.batch

        if not row.removed:
            row.removed = True

            if row.total_price:
                batch.total_price = (batch.total_price or 0) - row.total_price

        self.refresh_batch_status(batch)

    def execute(self, batch, user=None, progress=None, **kwargs):
        """
        Default behavior here will simply create a new (proper) Customer Order
        based on the batch contents.  Override as needed.
        """
        batch_fields = [
            'store',
            'id',
            'customer',
            'person',
            'phone_number',
            'email_address',
            'total_price',
        ]

        order = model.CustomerOrder()
        order.created_by = user
        order.status_code = self.enum.CUSTORDER_STATUS_ORDERED
        for field in batch_fields:
            setattr(order, field, getattr(batch, field))

        row_fields = [
            'product',
            'product_upc',
            'product_brand',
            'product_description',
            'product_size',
            'product_weighed',
            'department_number',
            'department_name',
            'case_quantity',
            'order_quantity',
            'order_uom',
            'product_unit_cost',
            'unit_price',
            'discount_percent',
            'total_price',
            'paid_amount',
            'payment_transaction_number',
        ]

        def convert(row, i):
            item = model.CustomerOrderItem()
            item.sequence = i + 1
            item.status_code = self.enum.CUSTORDER_ITEM_STATUS_ORDERED
            for field in row_fields:
                setattr(item, field, getattr(row, field))
            order.items.append(item)

        self.progress_loop(convert, batch.active_rows(), progress,
                           message="Converting batch rows to order items")

        session = orm.object_session(batch)
        session.add(order)
        session.flush()

        return order



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Excel utilities
"""

from __future__ import unicode_literals, absolute_import

import datetime

import six
import xlrd
from xlrd.xldate import xldate_as_tuple

import openpyxl
from openpyxl.styles import Font, PatternFill
from openpyxl.cell.cell import get_column_letter

from rattail.util import progress_loop


class ExcelReaderXLS(object):
    """
    Basic class for reading Excel "legacy" (.xls) files.

    Uses the ``xlrd`` package to read the files.
    """

    def __init__(self, path, sheet=0, sheet_name=None, header=0,
                 first_data_row=None,
                 datefmt='%Y-%m-%d', strip_fieldnames=True):
        """
        Constructor; opens an Excel file for reading.

        :param header: Which row should be used as the header, i.e. to
           determine field (column) names.  This is a zero-based index, so is 0
           by default (i.e. the first row).

        :param first_data_row: Which is the first row to contain data.  If not
           specified, it will be assumed that data rows begin immediately after
           the header row, as defined by :param:`header`.  This again is
           zero-based, so if the very first row is the true header, but then
           there is another "header" row also, you might specify a value of
           ``2`` here, since the 3rd row is the first to contain data.
        """
        self.book = xlrd.open_workbook(path)
        if sheet_name is not None:
            self.sheet = self.book.sheet_by_name(sheet_name)
        else:
            self.sheet = self.book.sheet_by_index(sheet)
        self.header = header
        if first_data_row is not None:
            self.first_data_row = first_data_row
        else:
            self.first_data_row = self.header + 1
        self.fields = self.sheet.row_values(self.header)
        if strip_fieldnames:
            self.fields = [field.strip() for field in self.fields]
        self.datefmt = datefmt

    def sheet_by_name(self, name):
        return self.book.sheet_by_name(name)

    def read_rows(self, progress=None):
        rows = []

        def append(row, i):
            values = self.sheet.row_values(row)
            data = dict([(self.fields[j], value)
                         for j, value in enumerate(values)])
            rows.append(data)

        progress_loop(append, range(self.first_data_row, self.sheet.nrows), progress,
                      message="Reading data from Excel file")
        return rows

    def parse_date(self, value, fmt=None):
        if isinstance(value, float):
            args = xldate_as_tuple(value, self.book.datemode)
            return datetime.datetime(*args).date()
        if value:
            return datetime.datetime.strptime(value, fmt or self.datefmt).date()

# TODO: this should become a base class and/or wrapper of some sort. for now
# the "default" reader assumes XLS since apparently that is all we have ever
# supported until now...which was a surprise to me...
ExcelReader = ExcelReaderXLS


class ExcelReaderXLSX(object):
    """
    Basic class for reading Excel 2010 (.xslx) files.

    Uses the ``openpyxl`` package to read the files.
    """

    def __init__(self, path, header_row=1, **kwargs):
        """
        Constructor; opens an Excel file for reading.

        :param path: Path to the Excel file.

        :param header_row: Which row contains the column headers.  This is
           1-based, so the 1 is the default (i.e. the first row).
        """
        self.book = openpyxl.load_workbook(filename=path)
        self.sheet = self.book.active

        self.header_row = header_row

        self.fields = None
        # TODO: this seems like a hacky way to get the header fields?  we are
        # "iterating" over the single header row, effectively...
        for row in self.sheet.iter_rows(min_row=self.header_row,
                                        max_row=self.header_row,
                                        values_only=True):
            assert self.fields is None
            self.fields = list(row)

    def read_rows(self, progress=None):
        rows = []

        def append(row, i):
            data = dict([(self.fields[j], value)
                         for j, value in enumerate(row)])
            rows.append(data)

        xlrows = list(self.sheet.iter_rows(min_row=self.header_row + 1,
                                           values_only=True))
        progress_loop(append, xlrows, progress,
                      message="Reading data from Excel file")
        return rows


class ExcelWriter(object):
    """
    Base class for Excel writers.
    """

    def __init__(self, path, fields, sheet_title=None, number_formats={}):
        """
        Constructor; opens an Excel workbook for writing.
        """
        self.path = path
        self.fields = fields
        self.book = openpyxl.Workbook()
        self.sheet = self.book.active
        if sheet_title:
            self.sheet.title = sheet_title
        self.number_formats = number_formats

    def create_sheet(self, title):
        """
        Create a new sheet in the workbook, and make it active.
        """
        self.sheet = self.book.create_sheet(title)
        return self.sheet

    def enable_grid_lines(self, sheet=None, enabled=True):
        if not sheet:
            sheet = self.sheet
        sheet.sheet_view.showGridLines = enabled

    def disable_grid_lines(self, sheet=None):
        self.enable_grid_lines(sheet=sheet, enabled=False)

    def write_header(self, labels=None):
        font = Font(bold=True)
        for i, field in enumerate(self.fields, 1):
            value = field
            if labels and field in labels:
                value = labels[field]
            cell = self.sheet.cell(row=1, column=i, value=value)
            cell.font = font

    def write_row(self, data, row=None):
        """
        Write (append) a single data row to the current sheet.

        :param row: The 1-based row number to which data should be written.
        """
        if row is None:
            raise NotImplementedError("should be able to detect 'next' row here?")

        self.sheet.append(data)

        # apply number formats
        if self.number_formats:
            for col, field in enumerate(self.fields, 1):
                if field in self.number_formats:
                    cell = self.sheet.cell(row=row, column=col)
                    cell.number_format = self.number_formats[field]

        # apply row highlighting
        if row % 2 == 0:
            fill_even = PatternFill(patternType='solid',
                                    fgColor='d9d9d9',
                                    bgColor='d9d9d9')
            for col, field in enumerate(self.fields, 1):
                cell = self.sheet.cell(row=row, column=col)
                cell.fill = fill_even

    def write_rows(self, rows, progress=None):
        """
        Write (append) a sequence of data rows to the current sheet.
        """
        def write(data, i):
            # must add 1 to account for header
            self.write_row(data, row=i + 1)

        progress_loop(write, rows, progress,
                      message="Generating cells for spreadsheet")

    def auto_freeze(self, row=2, column=1):
        """
        Freeze sheet per "the usual"
        """
        self.sheet.freeze_panes = self.sheet.cell(row=row, column=column)

    def auto_filter(self):
        """
        Add auto filters for all columns.
        """
        first = self.sheet.cell(row=1, column=1)
        last = self.sheet.cell(row=self.sheet.max_row, column=self.sheet.max_column)
        cellrange = '{}:{}'.format(first.coordinate, last.coordinate)
        self.sheet.auto_filter.ref = cellrange

    def auto_resize(self, progress=None):
        """
        (Try to) Auto-resize all data columns.
        """
        # note, some of the below uses efficiency tricks from these docs
        # https://openpyxl.readthedocs.io/en/stable/tutorial.html#accessing-many-cells

        # we must calculate desired column widths.  but for sake of progress,
        # we'll iterate through rows instead of columns, to do that.  (this is
        # just to give a higher total for the progress bar.)  so the first pass
        # is really just to cache all existing string widths.
        cached = []

        def cache(row, i):
            cached.append([len(six.text_type(value))
                           for value in row])

        progress_loop(cache, list(self.sheet.values), progress,
                      message="Calculating all string widths")

        # okay, now can determine ideal widths
        column_widths = []

        def calculate(col, i):
            width = max([row_widths[col] for row_widths in cached])
            column_widths.append(width or 5)

        progress_loop(calculate, range(self.sheet.max_column), progress,
                      message="Calculating desired column widths")

        # resize columns
        for i, width in enumerate(column_widths, 1):
            self.sheet.column_dimensions[get_column_letter(i)].width = width + 3

    def save(self, progress=None):
        """
        Save the Excel workbook to file.  If ``progress`` is provided, it will
        be used in a hacky sort of way, i.e. from 0 to 1 only since we have no
        way of knowing true progress for the save operation.  (But it can still
        be nice to let user know this is the step we're on at least.)
        """
        def save(x, i):
            self.book.save(self.path)

        progress_loop(save, range(1), progress,
                      message="Saving workbook to file")



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Database Stuff
"""

from __future__ import unicode_literals, absolute_import

try:
    import sqlalchemy
except ImportError:
    sqlalchemy = None
else:
    from sqlalchemy import orm
    from sqlalchemy.orm.exc import NoResultFound

from rattail.config import ConfigExtension as BaseExtension


if sqlalchemy:

    class SessionBase(orm.Session):
        """
        Custom SQLAlchemy session class, which adds some convenience methods
        related to the SQLAlchemy-Continuum integration.
        """

        def __init__(self, rattail_config=None, rattail_record_changes=None, continuum_user=None, **kwargs):
            """
            Custom constructor, to allow specifying the Continuum user at session
            creation.  If ``continuum_user`` is specified, its value will be passed
            to :meth:`set_continuum_user()`.
            """
            super(SessionBase, self).__init__(**kwargs)
            self.rattail_config = rattail_config

            # maybe record changes
            if rattail_record_changes is None:
                rattail_record_changes = getattr(self.bind, 'rattail_record_changes', False)
            if rattail_record_changes:
                from rattail.db.changes import record_changes
                record_changes(self, config=self.rattail_config)
            else:
                self.rattail_record_changes = False

            if continuum_user is None:
                self.continuum_user = None
            else:
                self.set_continuum_user(continuum_user)

        def set_continuum_user(self, user_info):
            """
            Set the effective Continuum user for the session.

            :param user_info: May be a :class:`model.User` instance, or the
              ``uuid`` or ``username`` for one.
            """
            from rattail.db import model

            if isinstance(user_info, model.User):
                user = self.merge(user_info)
            else:
                user = self.query(model.User).get(user_info)
                if not user:
                    try:
                        user = self.query(model.User).filter_by(username=user_info).one()
                    except NoResultFound:
                        user = None
            self.continuum_user = user


    Session = orm.sessionmaker(class_=SessionBase, rattail_config=None, expire_on_commit=False)


else: # no sqlalchemy
    Session = None


class ConfigExtension(BaseExtension):
    """
    Config extension for the ``rattail.db`` subpackage.  This extension is
    responsible for loading the available Rattail database engine(s), and
    configuring the :class:`Session` class with the default engine.  This
    extension expects to find something like the following in your config file:

    .. code-block:: ini

       [rattail.db]
       keys = default, host, other
       default.url = postgresql://localhost/rattail
       host.url = postgresql://host-server/rattail
       other.url = postgresql://other-server/rattail

    The result of this extension's processing is that the config object will
    get two new attributes:

    .. attribute:: rattail.config.RattailConfig.rattail_engines

       Dict of available Rattail database engines.  Keys of the dict are the
       same as found in the config file; values are the database engines.  Note
       that it is possible for this to be an empty dict.

    .. attribute:: rattail.config.RattailConfig.rattail_engine

       Default database engine; same as ``rattail_engines['default']``.  Note
       that it is possible for this value to be ``None``.
    """
    key = 'rattail.db'

    def configure(self, config):
        from rattail.db.config import get_engines, configure_session

        if Session:

            # Add Rattail database connection info to config.
            config.rattail_engines = get_engines(config)
            config.rattail_engine = config.rattail_engines.get('default')
            Session.configure(bind=config.rattail_engine, rattail_config=config)

            # TODO: This should be removed, it sets 'record changes' globally.
            configure_session(config, Session)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
People Handler

See also :doc:`rattail-manual:base/handlers/other/people`.
"""

from __future__ import unicode_literals, absolute_import

import warnings

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.app import GenericHandler
from rattail.time import make_utc


class PeopleHandler(GenericHandler):
    """
    Base class and default implementation for people handlers.
    """

    def normalize_full_name(self, first, last, **kwargs):
        """
        Normalize a "full" name based on the given first and last
        names.  Tries to be smart about collapsing whitespace etc.

        :param first: First name.
        :param last: Last name.
        :returns: First and last name combined.
        """
        from rattail.db.util import normalize_full_name
        return normalize_full_name(first, last)

    def make_person(self, **kwargs):
        """
        Make and return a new Person instance.
        """
        model = self.model
        person = model.Person()

        if 'first_name' in kwargs:
            person.first_name = kwargs.pop('first_name')
        if 'middle_name' in kwargs:
            person.middle_name = kwargs.pop('middle_name')
        if 'last_name' in kwargs:
            person.last_name = kwargs.pop('last_name')

        if 'display_name' in kwargs:
            person.display_name = kwargs.pop('display_name')
        else:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

        for key, value in six.iteritems(kwargs):
            if hasattr(person, key):
                setattr(person, key, value)

        return person

    def update_names(self, person, **kwargs):
        """
        Update name(s) for the given person.

        :param person: Reference to a ``Person`` record.
        :param first: First name for the person.
        :param middle: Middle name for the person.
        :param last: Last name for the person.
        :param full: Full (display) name for the person.
        """
        if 'first' in kwargs:
            person.first_name = kwargs['first']

        if 'middle' in kwargs:
            person.middle_name = kwargs['middle']

        if 'last' in kwargs:
            person.last_name = kwargs['last']

        if 'full' in kwargs:
            if kwargs['full']:
                person.display_name = kwargs['full']
            else:
                person.display_name = self.normalize_full_name(
                    person.first_name, person.last_name)
        elif 'first' in kwargs and 'last' in kwargs:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

    def ensure_address(self, person, **kwargs):
        """
        Returns the default address record associated with the given
        person, creating it first if necessary.
        """
        address = person.first_address()
        if not address:
            address = person.add_address(**kwargs)
            # TODO: this might be a good idea..maybe if a kwarg flag is set?
            # person.set_primary_address(address)
        return address

    def get_merge_preview_fields(self, **kwargs):
        """
        Returns a sequence of fields which will be used during a merge
        preview.
        """
        def F(name, **kwargs):
            field = {'name': name}
            field.update(kwargs)
            return field

        return [
            F('uuid'),
            F('first_name'),
            F('last_name'),
            F('display_name'),
            F('usernames', additive=True),
            F('member_uuids', additive=True),
        ]

    def get_merge_preview_data(self, person, **kwargs):
        """
        Must return a data dictionary for the given person, which can
        be presented to the user during a merge preview.
        """
        return {
            'uuid': person.uuid,
            'first_name': person.first_name,
            'last_name': person.last_name,
            'display_name': person.display_name,
            'usernames': [u.username for u in person.users],
            'member_uuids': [m.uuid for m in person.members],
        }

    def get_merge_resulting_data(self, removing, keeping, **kwargs):
        """
        Must return a dictionary to represent what the *final* data
        would look like, should the proposed merge occur.  Note that
        we're still in preview mode here, this doesn't actually cause
        any particular data to become final.

        :param removing: Data dictionary for the Person to be removed,
           as obtained via :meth:`get_merge_preview_data()`.
        :param keeping: Data dictionary for the Person to be preserved,
           as obtained via :meth:`get_merge_preview_data()`.
        """
        fields = self.get_merge_preview_fields()
        coalesce_fields = [f for f in fields
                           if f.get('coalesce')]
        additive_fields = [f for f in fields
                           if f.get('additive')]

        # start with clone of the `keeping` dict
        result = dict(keeping)

        # coalesce any field values which need it
        for field in coalesce_fields:
            if removing[field] is not None and keeping[field] is None:
                result[field] = removing[field]
            elif removing[field] and not keeping[field]:
                result[field] = removing[field]

        # sum any field values which need it
        for field in additive_fields:
            if isinstance(keeping[field], (list, tuple)):
                result[field] = sorted(set(removing[field] + keeping[field]))
            else:
                result[field] = removing[field] + keeping[field]

        return result

    def why_not_merge(self, removing, keeping, **kwargs):
        """
        Evaluate the given merge candidates and if there is a reason *not*
        to merge them, return that reason.

        :param removing: Person record which will be removed, should the
           merge happen.
        :param keeping: Person record which will be kept, should the
           merge happen.
        :returns: String indicating reason not to merge, or ``None``.
        """

    def perform_merge(self, removing, keeping, **kwargs):
        """
        Perform an actual merge of the 2 given people.

        :param removing: Person record which should be removed.
        :param keeping: Person record which should be kept.
        """
        # move Member records to final Person
        for member in list(removing.members):
            removing.members.remove(member)
            keeping.members.append(member)

        # move User records to final Person
        for user in list(removing.users):
            removing.users.remove(user)
            keeping.users.append(user)

        # delete unwanted Person
        session = orm.object_session(keeping)
        session.delete(removing)
        session.flush()

        self.satisfy_merge_requests(removing, keeping, user=kwargs.get('user'))

    def satisfy_merge_requests(self, removing, keeping, user):
        """
        If there was a merge request(s) for this pair, mark it complete.
        """
        session = orm.object_session(keeping)
        model = self.model
        merge_requests = session.query(model.MergePeopleRequest)\
                                .filter(sa.or_(
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == removing.uuid,
                                        model.MergePeopleRequest.keeping_uuid == keeping.uuid),
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == keeping.uuid,
                                        model.MergePeopleRequest.keeping_uuid == removing.uuid)))\
                                .all()
        for merge_request in merge_requests:
            # set the record straight re: removing vs. keeping
            merge_request.removing_uuid = removing.uuid
            merge_request.keeping_uuid = keeping.uuid
            merge_request.merged = make_utc()
            merge_request.merged_by = user

def get_people_handler(config, **kwargs):
    """
    Create and return the configured :class:`PeopleHandler` instance.

    .. warning::
       This function is deprecated; please use
       :meth:`~rattail.app.AppHandler.get_people_handler` instead.
    """
    warnings.warn("get_people_handler() function is deprecated, "
                  "please use app.get_people_handler() method instead",
                  DeprecationWarning)
    app = config.get_app()
    return app.get_people_handler(**kwargs)



# -*- coding: utf-8 -*-
"""add new versioning

Revision ID: 3b2859694c53
Revises: e9fca3d787ea
Create Date: 2017-05-25 18:28:45.492796

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '3b2859694c53'
down_revision = u'e9fca3d787ea'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # transaction
    op.create_table('transaction',
                    sa.Column('issued_at', sa.DateTime(), nullable=True),
                    sa.Column('id', sa.BigInteger(), nullable=False),
                    sa.Column('remote_addr', sa.String(length=50), nullable=True),
                    sa.Column('user_id', sa.String(length=32), nullable=True),
                    sa.ForeignKeyConstraint(['user_id'], [u'user.uuid'], ),
                    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_transaction_user_id'), 'transaction', ['user_id'], unique=False)

    # person_version
    op.create_table('person_version',
                    sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
                    sa.Column('first_name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('middle_name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('last_name', sa.String(length=50), autoincrement=False, nullable=True),
                    sa.Column('display_name', sa.String(length=100), autoincrement=False, nullable=True),
                    sa.Column('transaction_id', sa.BigInteger(), autoincrement=False, nullable=False),
                    sa.Column('end_transaction_id', sa.BigInteger(), nullable=True),
                    sa.Column('operation_type', sa.SmallInteger(), nullable=False),
                    sa.PrimaryKeyConstraint('uuid', 'transaction_id')
    )
    op.create_index(op.f('ix_person_version_end_transaction_id'), 'person_version', ['end_transaction_id'], unique=False)
    op.create_index(op.f('ix_person_version_operation_type'), 'person_version', ['operation_type'], unique=False)
    op.create_index(op.f('ix_person_version_transaction_id'), 'person_version', ['transaction_id'], unique=False)


def downgrade():

    # person
    op.drop_index(op.f('ix_person_version_transaction_id'), table_name='person_version')
    op.drop_index(op.f('ix_person_version_operation_type'), table_name='person_version')
    op.drop_index(op.f('ix_person_version_end_transaction_id'), table_name='person_version')
    op.drop_table('person_version')

    # transaction
    op.drop_index(op.f('ix_transaction_user_id'), table_name='transaction')
    op.drop_table('transaction')



# -*- coding: utf-8 -*-
"""add employee worked shifts

Revision ID: 017daa26c36c
Revises: 33c3b9acc341
Create Date: 2016-04-25 14:58:32.182768

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '017daa26c36c'
down_revision = u'33c3b9acc341'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # employee_shift_worked
    op.create_table('employee_shift_worked',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('employee_uuid', sa.String(length=32), nullable=False),
                    sa.Column('store_uuid', sa.String(length=32), nullable=True),
                    sa.Column('punch_in', sa.DateTime(), nullable=True),
                    sa.Column('punch_out', sa.DateTime(), nullable=True),
                    sa.ForeignKeyConstraint(['employee_uuid'], [u'employee.uuid'], name=u'employee_shift_worked_fk_employee'),
                    sa.ForeignKeyConstraint(['store_uuid'], [u'store.uuid'], name=u'employee_shift_worked_fk_store'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # employee_shift_worked
    op.drop_table('employee_shift_worked')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Email Bouncer Configuration
"""

from __future__ import unicode_literals

from rattail.config import parse_list
from rattail.util import load_object
from rattail.exceptions import ConfigurationError


class Profile(object):
    """
    Simple class to hold configuration for an email bouncer "profile".  Each
    profile determines how to connect to an IMAP folder, which handler to
    invoke when new messages appear, etc.  Each instance of this class has the
    following attributes:

    .. attribute:: config

       Reference to the underlying configuration object from which the profile
       derives its other attributes.

    .. attribute:: key

       String which differentiates this profile from any others which may exist
       within the configuration.

    .. attribute:: imap_server

       Hostname or IP address of the IMAP server to which to connect.

    .. attribute:: imap_username

       Username to use when logging into the IMAP server.

    .. attribute:: imap_password

       Password to use when logging into the IMAP server.

    .. attribute:: imap_folder

       Folder to select when checking the IMAP server for new messages.

    .. attribute:: handler_spec

       Spec string for the handler.  This must be a subclass of
       :class:`rattail.bouncer:BounceHandler`.

    .. attribute:: handler

       Reference to the handler instance.  This will be a subclass of
       :class:`rattail.bouncer:BounceHandler`.
    """

    def __init__(self, config, key):
        self.config = config
        self.key = key

        self.imap_server = self._config_string('imap.server')
        self.imap_username = self._config_string('imap.username')
        self.imap_password = self._config_string('imap.password')
        self.imap_folder_inbox = self._config_string('imap.inbox')
        self.imap_folder_backup = self._config_string('imap.backup')
        self.imap_delay = self._config_int('imap.delay', default=120)
        self.imap_recycle = self._config_int('imap.recycle', default=1200)

        self.workdir = self._config_string('workdir')

        self.handler_spec = self._config_string('handler') or 'rattail.bouncer:BounceHandler'
        self.handler = load_object(self.handler_spec)(config, key)

    def _config_string(self, option):
        return self.config.get('rattail.bouncer', '{0}.{1}'.format(self.key, option))

    def _config_int(self, option, minimum=1, default=None):
        option = '{0}.{1}'.format(self.key, option)
        if self.config.has_option('rattail.bouncer', option):
            value = self.config.getint('rattail.bouncer', option)
            if value < minimum:
                log.warning("config value {0} is too small; falling back to minimum "
                            "of {1} for option: {2}".format(value, minimum, option))
                value = minimum
        elif default is not None and default >= minimum:
            value = default
        else:
            value = minimum
        return value


def get_profile_keys(config):
    """
    Returns a list of keys used in the bouncer configuration.
    """
    keys = config.get('rattail.bouncer', 'watch')
    if keys:
        return parse_list(keys)


def load_profiles(config):
    """
    Load all active email bouncer profiles defined within configuration.
    """
    # Make sure we have a top-level directive.
    keys = get_profile_keys(config)
    if not keys:
        raise ConfigurationError(
            "The bouncer configuration does not specify any IMAP profiles to "
            "be watched.  Please defined the 'watch' option within the "
            "[rattail.bouncer] section of your config file.")

    watched = {}
    for key in keys:
        profile = Profile(config, key)
        watched[key] = profile
    return watched



# -*- coding: utf-8; -*-
"""grow role.name

Revision ID: e1edeba632f8
Revises: 11a9145414a9
Create Date: 2018-11-13 10:46:53.014681

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'e1edeba632f8'
down_revision = '11a9145414a9'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # role
    op.alter_column('role', 'name', type_=sa.String(length=100))


def downgrade():

    # role
    op.alter_column('role', 'name', type_=sa.String(length=25))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor invoice parser for KeHE Distributors
"""

from __future__ import unicode_literals, absolute_import

import re
import csv
import datetime

import six

from rattail.db import model
from rattail.gpc import GPC
from rattail.vendors.invoices import InvoiceParser


class KeheInvoiceParser(InvoiceParser):
    """
    Vendor invoice parser for KeHE Distributors.

    This parser is actually capable of handling two different but similar
    formats.  CSV data is expected in both cases.  Basically the KeHE portal
    changed the format available for download, so the parser now needs to
    support both the "old" as well as "new" formats.  Note that the newer
    formats appeared circa 2020-02-11.

    See the :meth:`detect_version()` method for details of how the parser
    decides which "version" (format) of file it's dealing with.
    """
    key = 'rattail.contrib.kehe'
    display = "KeHE Distributors"
    vendor_key = 'kehe'

    pack_size_pattern = re.compile('^(?P<case_quantity>\d+)/(?P<size>\d*\.\d+ \w\w)$')

    def detect_version(self, path):
        """
        This will inspect the data file and return the "version" of format it
        thinks we're dealing with.
        """
        with open(path, 'rt') as f:
            line = f.readline()

        # if line has no tab characters, we can safely assume new version 2
        if '\t' not in line:
            return 2

        # the file *does* have tab characters, so we must check a column header
        if six.PY3:
            csv_file = open(path, 'rt')
            reader = csv.DictReader(csv_file, delimiter='\t')
        else: # PY2
            csv_file = open(path, 'rb')
            reader = csv.DictReader(csv_file, delimiter=b'\t')
        data = next(reader)
        csv_file.close()

        # check for old version 1 column header
        if 'Invoice Date' in data:
            return 1

        # okay, assume new version 3 then
        # assert 'InvoiceDate' in data
        return 3

    def parse_invoice_date(self, path):

        # first find out which version of file we have
        version = self.detect_version(path)

        delimiter = str('\t') if version in (1, 3) else ','
        if six.PY3:
            csv_file = open(path, 'rt')
            reader = csv.DictReader(csv_file, delimiter=delimiter)
        else: # PY2
            csv_file = open(path, 'rb')
            reader = csv.DictReader(csv_file, delimiter=delimiter)
        data = next(reader)
        csv_file.close()

        if version == 1:
            return datetime.datetime.strptime(data['Invoice Date'], '%Y-%m-%d').date()
        else: # version in (2, 3)
            return datetime.datetime.strptime(data['InvoiceDate'], '%m/%d/%Y %I:%M:%S %p').date()

    def parse_rows(self, path):

        # first find out which version of file we have
        version = self.detect_version(path)
        delimiter = str('\t') if version in (1, 3) else ','

        # default fields for old version 1
        fields = {
            'upc': 'UPC Code',
            'ship_item': 'Ship Item',
            'brand': 'Brand',
            'description': 'Description',
            'order_quantity': 'Order Qty',
            'ship_quantity': 'Ship Qty',
            'net_each': 'Net Each',
            'net_billable': 'Net Billable',
            'pack_size': 'Pack Size',
        }

        if version in (2, 3):
            fields.update({
                'upc': 'Upc',
                'ship_item': 'ShipItem',
                'order_quantity': 'OrderQuantity',
                'ship_quantity': 'ShipQuantity',
                'net_each': 'NetEach',
                'net_billable': 'NetBillable',
                'pack_size': 'PackSize',
            })

        if six.PY3:
            csv_file = open(path, 'rt')
            reader = csv.DictReader(csv_file, delimiter=delimiter)
        else: # PY2
            csv_file = open(path, 'rb')
            reader = csv.DictReader(csv_file, delimiter=delimiter)

        for data in reader:

            row = model.VendorInvoiceBatchRow()
            row.item_entry = data[fields['upc']]
            row.upc = GPC(row.item_entry)
            row.vendor_code = data[fields['ship_item']]
            row.brand_name = data[fields['brand']]
            row.description = data[fields['description']]
            row.ordered_units = self.int_(data[fields['order_quantity']])
            row.shipped_units = self.int_(data[fields['ship_quantity']])
            row.unit_cost = self.decimal(data[fields['net_each']])
            row.total_cost = self.decimal(data[fields['net_billable']])

            # Case quantity may be embedded in size string.
            row.size = data[fields['pack_size']]
            row.case_quantity = 1
            match = self.pack_size_pattern.match(row.size)
            if match:
                row.case_quantity = int(match.group('case_quantity'))
                row.size = match.group('size')

            # attach original raw data to the row we're returning; caller
            # can use as needed, or ignore
            row._raw_data = data

            yield row

        csv_file.close()



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

from unittest import TestCase

import six
from mock import patch, Mock

import sqlalchemy as sa
from sqlalchemy import orm
from sqlalchemy.orm.exc import MultipleResultsFound

from rattail.importing import sqlalchemy as saimport


class Widget(object):
    pass

metadata = sa.MetaData()

widget_table = sa.Table('widgets', metadata,
                        sa.Column('id', sa.Integer(), primary_key=True),
                        sa.Column('description', sa.String(length=50)))

widget_mapper = orm.mapper(Widget, widget_table)

WIDGETS = [
    {'id': 1, 'description': "Main Widget"},
    {'id': 2, 'description': "Other Widget"},
    {'id': 3, 'description': "Other Widget"},
]


class TestFromSQLAlchemy(TestCase):

    def test_query(self):
        Session = orm.sessionmaker(bind=sa.create_engine('sqlite://'))
        session = Session()
        importer = saimport.FromSQLAlchemy(host_session=session,
                                           host_model_class=Widget)
        self.assertEqual(six.text_type(importer.query()),
                         "SELECT widgets.id AS widgets_id, widgets.description AS widgets_description \n"
                         "FROM widgets")


class TestToSQLAlchemy(TestCase):

    def setUp(self):
        engine = sa.create_engine('sqlite://')
        metadata.create_all(bind=engine)
        Session = orm.sessionmaker(bind=engine)
        self.session = Session()
        for data in WIDGETS:
            widget = Widget()
            for key, value in data.items():
                setattr(widget, key, value)
            self.session.add(widget)
        self.session.commit()

    def tearDown(self):
        self.session.close()
        del self.session

    def make_importer(self, **kwargs):
        kwargs.setdefault('model_class', Widget)
        return saimport.ToSQLAlchemy(**kwargs)

    def test_model_mapper(self):
        importer = self.make_importer()
        self.assertIs(importer.model_mapper, widget_mapper)

    def test_model_table(self):
        importer = self.make_importer()
        self.assertIs(importer.model_table, widget_table)

    def test_simple_fields(self):
        importer = self.make_importer()
        self.assertEqual(importer.simple_fields, ['id', 'description'])

    def test_get_single_local_object(self):

        # simple key
        importer = self.make_importer(key='id', session=self.session)
        widget = importer.get_single_local_object((1,))
        self.assertEqual(widget.id, 1)
        self.assertEqual(widget.description, "Main Widget")

        # compound key
        importer = self.make_importer(key=('id', 'description'), session=self.session)
        widget = importer.get_single_local_object((1, "Main Widget"))
        self.assertEqual(widget.id, 1)
        self.assertEqual(widget.description, "Main Widget")

        # widget not found
        importer = self.make_importer(key='id', session=self.session)
        self.assertIsNone(importer.get_single_local_object((42,)))

        # multiple widgets found
        importer = self.make_importer(key='description', session=self.session)
        self.assertRaises(MultipleResultsFound, importer.get_single_local_object, ("Other Widget",))

    def test_create_object(self):
        importer = self.make_importer(key='id', session=self.session)
        widget = importer.create_object((42,), {'id': 42, 'description': "Latest Widget"})
        self.assertIn(widget, self.session)
        self.assertIn(widget, self.session.new)
        self.assertEqual(widget.id, 42)
        self.assertEqual(widget.description, "Latest Widget")

    def test_delete_object(self):
        widget = self.session.query(Widget).get(1)
        self.assertIn(widget, self.session)
        importer = self.make_importer(session=self.session)
        self.assertTrue(importer.delete_object(widget))
        self.assertIn(widget, self.session)
        self.assertIn(widget, self.session.deleted)
        self.session.flush()
        self.assertIsNone(self.session.query(Widget).get(1))

    def test_cache_model(self):
        importer = self.make_importer(key='id', session=self.session)
        cache = importer.cache_model(Widget, key='id')
        self.assertEqual(len(cache), 3)
        for i in range(1, 4):
            self.assertIn(i, cache)
            self.assertIsInstance(cache[i], Widget)
            self.assertEqual(cache[i].id, i)
            self.assertEqual(cache[i].description, WIDGETS[i-1]['description'])

    def test_cache_local_data(self):
        importer = self.make_importer(key='id', session=self.session)
        importer.handler = Mock(local_title="Nevermind")
        cache = importer.cache_local_data()
        self.assertEqual(len(cache), 3)
        for i in range(1, 4):
            self.assertIn((i,), cache)
            cached = cache[(i,)]
            self.assertIsInstance(cached, dict)
            self.assertIsInstance(cached['object'], Widget)
            self.assertEqual(cached['object'].id, i)
            self.assertEqual(cached['object'].description, WIDGETS[i-1]['description'])
            self.assertIsInstance(cached['data'], dict)
            self.assertEqual(cached['data']['id'], i)
            self.assertEqual(cached['data']['description'], WIDGETS[i-1]['description'])

    # TODO: lame
    def test_flush_session(self):
        importer = self.make_importer(fields=['id'], key='id', session=self.session, flush_session=True)
        widget = Widget()
        widget.id = 1
        widget, original = importer.update_object(widget, {'id': 1}), widget
        self.assertIs(widget, original)

    def test_flush_create_update(self):
        importer = saimport.ToSQLAlchemy(fields=['id'], session=self.session)
        with patch.object(self.session, 'flush') as flush:
            importer.flush_create_update()
            flush.assert_called_once_with()



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import shutil
import tempfile
import unittest

import lockfile

from rattail import files


class TestLockingCopy(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

        self.srcdir = os.path.join(self.tempdir, 'src')
        os.makedirs(self.srcdir)

        self.dstdir = os.path.join(self.tempdir, 'dst')
        os.makedirs(self.dstdir)

        self.src_file = os.path.join(self.srcdir, 'somefile')
        with open(self.src_file, 'wt') as f:
            f.write('')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test_normal_copy_succeeds(self):
        files.locking_copy(self.src_file, self.dstdir)
        dst_file = os.path.join(self.dstdir, 'somefile')
        self.assertTrue(os.path.exists(dst_file))
        self.assertTrue(os.path.isfile(dst_file))
        self.assertFalse(os.path.exists(os.path.join(self.dstdir, 'somefile.lock')))



# -*- coding: utf-8 -*-
"""add inventory batch mode

Revision ID: f5c936a4cc62
Revises: 1f8e3cf6b2a2
Create Date: 2016-08-22 21:41:25.359322

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = 'f5c936a4cc62'
down_revision = u'1f8e3cf6b2a2'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


batch_id_seq = sa.Sequence('batch_id_seq')


def upgrade():

    # batch_inventory
    op.add_column('batch_inventory', sa.Column('mode', sa.Integer(), nullable=True))


def downgrade():

    # batch_inventory
    op.drop_column('batch_inventory', 'mode')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Model for handheld batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.db.model import Base, FileBatchMixin, ProductBatchRowMixin


class HandheldBatch(FileBatchMixin, Base):
    """
    Batch representing a file from a batch-oriented handheld device
    (e.g. Motorola).
    """
    __tablename__ = 'batch_handheld'
    __batchrow_class__ = 'HandheldBatchRow'
    batch_key = 'handheld'

    # TODO: this should probably be new default
    filename_nullable = True

    device_type = sa.Column(sa.String(length=50), nullable=True, doc="""
    Key indicating the type of device where the batch originated.
    """)

    device_name = sa.Column(sa.String(length=50), nullable=True, doc="""
    Name of the device where the batch originated.
    """)


class HandheldBatchRow(ProductBatchRowMixin, Base):
    """
    Rows for handheld batches.
    """
    __tablename__ = 'batch_handheld_row'
    __batch_class__ = HandheldBatch

    STATUS_OK = 1
    STATUS_PRODUCT_NOT_FOUND = 2

    STATUS = {
        STATUS_OK:                      "ok",
        STATUS_PRODUCT_NOT_FOUND:       "product not found",
    }

    cases = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Case quantity for the record.
    """)

    units = sa.Column(sa.Numeric(precision=10, scale=4), nullable=True, doc="""
    Unit quantity for the record.
    """)



# -*- coding: utf-8; -*-
"""add product inventory

Revision ID: 69d924015613
Revises: a28c3583c8f0
Create Date: 2017-07-18 21:27:41.750784

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '69d924015613'
down_revision = u'a28c3583c8f0'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product_inventory
    op.create_table('product_inventory',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('product_uuid', sa.String(length=32), nullable=False),
                    sa.Column('on_hand', sa.Numeric(precision=9, scale=4), nullable=True),
                    sa.Column('on_order', sa.Numeric(precision=9, scale=4), nullable=True),
                    sa.ForeignKeyConstraint(['product_uuid'], [u'product.uuid'], name=u'product_inventory_fk_product'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # product_inventory
    op.drop_table('product_inventory')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Clientele Handler
"""

from __future__ import unicode_literals, absolute_import

from rattail.util import load_object
from rattail.app import GenericHandler


class ClienteleHandler(GenericHandler):
    """
    Base class and default implementation for clientele handlers.
    """

    def ensure_customer(self, person):
        """
        Returns the customer record associated with the given person, creating
        it first if necessary.
        """
        customer = self.get_customer(person)
        if customer:
            return customer

        session = self.get_session(person)
        customer = self.make_customer(person)
        session.add(customer)
        session.flush()
        session.refresh(person)
        return customer

    def get_customer(self, person):
        """
        Returns the customer associated with the given person, if there is one.
        """
        return person.only_customer(require=False)

    def get_person(self, customer):
        """
        Returns the person associated with the given customer, if there is one.
        """
        return customer.only_person(require=False)

    def make_customer(self, person, **kwargs):
        """
        Create and return a new customer record.
        """
        customer = self.model.Customer()
        customer.name = person.display_name
        customer.people.append(person)
        return customer


def get_clientele_handler(config, **kwargs):
    """
    Create and return the configured :class:`ClienteleHandler` instance.
    """
    spec = config.get('rattail', 'clientele.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = ClienteleHandler
    return factory(config, **kwargs)



# -*- coding: utf-8 -*-
"""improve size fields

Revision ID: 3544687ae150
Revises: ee253b66b7b
Create Date: 2015-02-26 23:10:16.777511

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3544687ae150'
down_revision = u'ee253b66b7b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from rattail import enum


def upgrade():

    # product.unit_of_measure
    product = sa.sql.table('product', sa.sql.column('unit_of_measure'))
    op.execute(product.update()\
                   .where(product.c.unit_of_measure == None)\
                   .values({'unit_of_measure': enum.UNIT_OF_MEASURE_NONE}))
    product_version = sa.sql.table('product_version', sa.sql.column('unit_of_measure'))
    op.execute(product_version.update()\
                   .where(product_version.c.unit_of_measure == None)\
                   .values({'unit_of_measure': enum.UNIT_OF_MEASURE_NONE}))
    op.alter_column('product', 'unit_of_measure', nullable=False)

    # product.unit_size
    op.add_column('product', sa.Column('unit_size', sa.Numeric(precision=8, scale=3), nullable=True))
    op.add_column('product_version', sa.Column('unit_size', sa.Numeric(precision=8, scale=3), nullable=True))

    # product.weighed
    op.add_column('product', sa.Column('weighed', sa.Boolean(), nullable=True))
    product = sa.sql.table('product', sa.sql.column('weighed'))
    op.execute(product.update().values({'weighed': False}))
    op.alter_column('product', 'weighed', nullable=False)
    op.add_column('product_version', sa.Column('weighed', sa.Boolean(), nullable=True))


def downgrade():

    # product.weighed
    op.drop_column('product_version', 'weighed')
    op.drop_column('product', 'weighed')

    # product.unit_size
    op.drop_column('product_version', 'unit_size')
    op.drop_column('product', 'unit_size')

    # product.unit_of_measure
    op.alter_column('product', 'unit_of_measure', nullable=True)



# -*- coding: utf-8 -*-
"""add change uuid

Revision ID: 835215361bb
Revises: 188810a5f9db
Create Date: 2015-10-22 23:19:44.460642

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '835215361bb'
down_revision = u'188810a5f9db'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # change
    op.drop_table('change')
    op.create_table(
        'change',
        sa.Column('uuid', sa.String(length=32), nullable=False),
        sa.Column('class_name', sa.String(length=25), nullable=False),
        sa.Column('instance_uuid', sa.String(length=32), nullable=False),
        sa.Column('deleted', sa.Boolean(), nullable=False),
        sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # change
    op.drop_table('change')
    op.create_table(
        'change',
        sa.Column('class_name', sa.String(length=25), autoincrement=False, nullable=False),
        sa.Column('uuid', sa.String(length=32), autoincrement=False, nullable=False),
        sa.Column('deleted', sa.Boolean(), autoincrement=False, nullable=True),
        sa.PrimaryKeyConstraint('class_name', 'uuid', name=u'changes_pkey')
    )



# -*- coding: utf-8; -*-
"""grow product.uom_abbreviation

Revision ID: 45b90e4f8ab4
Revises: c5f39e2dec78
Create Date: 2018-12-19 13:00:48.684227

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '45b90e4f8ab4'
down_revision = u'c5f39e2dec78'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # product
    op.alter_column('product', 'uom_abbreviation', type_=sa.String(length=10))
    op.alter_column('product_version', 'uom_abbreviation', type_=sa.String(length=10))


def downgrade():

    # product
    op.alter_column('product_version', 'uom_abbreviation', type_=sa.String(length=4))
    op.alter_column('product', 'uom_abbreviation', type_=sa.String(length=4))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data Models for Organization etc.
"""

from __future__ import unicode_literals, absolute_import

import six
import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db.model import Base, uuid_column


@six.python_2_unicode_compatible
class Department(Base):
    """
    Represents an organizational department.
    """
    __tablename__ = 'department'
    __versioned__ = {}

    uuid = uuid_column()

    number = sa.Column(sa.Integer(), nullable=True, doc="""
    Presumably unique number for the department, as integer.
    """)

    name = sa.Column(sa.String(length=30), nullable=True, doc="""
    Name for the department.
    """)

    product = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the department may be used for product
    organization, irrespective of its :attr:`personnel` flag.
    """)

    personnel = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether the department may be used for personnel
    organization, irrespective of its :attr:`product` flag.
    """)

    exempt_from_gross_sales = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating whether items in this department should (not) contribute to
    gross sales figures.  The SIL book says this about it:

       Dollar sales are not to be included in gross sales. (i.e., Postage
       Stamps, Lottery Tickets, etc.).
    """)

    allow_product_deletions = sa.Column(sa.Boolean(), nullable=True, doc="""
    Whether products belonging to this department may be deleted.
    """)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Subdepartment(Base):
    """
    Represents an organizational subdepartment.
    """
    __tablename__ = 'subdepartment'
    __table_args__ = (
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='subdepartment_fk_department'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    number = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=30))
    department_uuid = sa.Column(sa.String(length=32))

    def __str__(self):
        return self.name or ''


Department.subdepartments = orm.relationship(
    Subdepartment,
    backref='department',
    order_by=Subdepartment.name)


@six.python_2_unicode_compatible
class Category(Base):
    """
    Represents an organizational category for products.
    """
    __tablename__ = 'category'
    __table_args__ = (
        sa.ForeignKeyConstraint(['department_uuid'], ['department.uuid'], name='category_fk_department'),
        sa.UniqueConstraint('code', name='category_uq_code'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    code = sa.Column(sa.String(length=20), nullable=True, doc="""
    Unique code for the category, as string.
    """)

    number = sa.Column(sa.Integer(), nullable=True, doc="""
    Number for the category.  NOTE: Most code leverages the :attr:`code`
    instead of this number, which is essentially ignored.
    """)

    name = sa.Column(sa.String(length=50), nullable=True, doc="""
    Descriptive name for the category.
    """)

    department_uuid = sa.Column(sa.String(length=32))
    department = orm.relationship(Department, doc="""
    Reference to the :class:`Department` to which this category belongs, if
    applicable.
    """)

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class Family(Base):
    """
    Represents an organizational family for products.
    """
    __tablename__ = 'family'
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer())
    name = sa.Column(sa.String(length=50))

    def __str__(self):
        return self.name or ''


@six.python_2_unicode_compatible
class ReportCode(Base):
    """
    Represents an organizational "report code" for products.
    """
    __tablename__ = 'report_code'
    __table_args__ = (
        sa.UniqueConstraint('code', name='report_code_uq_code'),
    )
    __versioned__ = {}

    uuid = uuid_column()
    code = sa.Column(sa.Integer(), nullable=False)
    name = sa.Column(sa.String(length=100), nullable=True)

    def __str__(self):
        if not self.code:
            return ''
        return "{} - {}".format(self.code, self.name or '')


@six.python_2_unicode_compatible
class DepositLink(Base):
    """
    Represents a bottle or similar deposit, to which an item may be linked.
    """
    __tablename__ = 'deposit_link'
    __table_args__ = (
        sa.UniqueConstraint('code', name='deposit_link_uq_code'),
    )
    __versioned__ = {}

    uuid = uuid_column()

    code = sa.Column(sa.String(length=20), nullable=False, doc="""
    Unique "code" for the deposit link.
    """)

    description = sa.Column(sa.String(length=255), nullable=True)
    amount = sa.Column(sa.Numeric(precision=5, scale=2), nullable=False)

    def __str__(self):
        if self.description:
            return self.description
        return "${:0.2f}".format(self.amount)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Products Handler
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm

from rattail import pod
from rattail.util import load_object
from rattail.app import GenericHandler


class ProductsHandler(GenericHandler):
    """
    Base class and default implementation for product handlers.

    A products handler of course should get the final say in how products are
    handled.  This means everything from pricing, to whether or not a
    particular product can be deleted, etc.
    """

    def get_image_url(self, upc=None, product=None, **kwargs):
        """
        Return the preferred image URL for the given UPC or product.
        """
        if product and product.upc:
            upc = product.upc
        if upc:
            return self.get_pod_image_url(upc)

    def get_pod_image_url(self, upc, **kwargs):
        """
        Return the POD image URL for the given UPC.
        """
        if upc:
            return pod.get_image_url(self.config, upc)

    def get_uom_sil_codes(self, session, uppercase=False, **kwargs):
        """
        This should return a dict, keys of which are UOM abbreviation strings,
        and values of which are corresponding SIL code strings.

        :param session: Reference to current Rattail DB session.
        :param uppercase: Set to ``True`` to cause all UOM abbreviations to be
           upper-cased when adding to the map.
        :returns: Dictionary containing all known UOM / SIL code mappings.
        """
        model = self.model

        def normalize(uom):
            if uom.sil_code:
                return uom.sil_code

        def make_key(uom, normal):
            key = uom.abbreviation
            if uppercase:
                key = key.upper()
            return key

        return self.cache_model(session, model.UnitOfMeasure,
                                normalizer=normalize,
                                key=make_key)

    def get_uom_sil_code(self, session, uom, uppercase=False, **kwargs):
        """
        This should return a SIL code which corresponds to the given UOM
        abbreviation string.  Useful when you just need one out of the blue,
        but if you need multiple codes looked up then you're probably better
        off using :meth:`get_uom_sil_codes()` for efficiency.

        :param session: Reference to current Rattail DB session.
        :param uppercase: Set to ``True`` to cause the UOM abbreviation to be
           upper-cased before performing the lookup.  This effectively makes
           the search case-insensitive.
        :param uom:  Unit of measure as abbreviated string, e.g. ``'LB'``.
        :returns: SIL code for the UOM, as string (e.g. ``'49'``), or ``None``
           if no matching code was found.
        """
        model = self.model
        query = session.query(model.UnitOfMeasure)
        if uppercase:
            query = query.filter(sa.func.upper(model.UnitOfMeasure.abbreviation) == uom.upper())
        else:
            query = query.filter(model.UnitOfMeasure.abbreviation == uom)
        try:
            match = query.one()
        except orm.exc.NoResultFound:
            pass
        else:
            return match.sil_code

    def collect_wild_uoms(self, **kwargs):
        """
        Collect all UOM abbreviations "from the wild" and ensure each is
        represented within the Rattail Units of Measure table.

        Note that you should not need to override this method.  Please override
        :meth:`find_wild_uoms()` instead.
        """
        session = self.make_session()
        model = self.model

        wild_uoms = self.find_wild_uoms(session, **kwargs)

        known_uoms = self.cache_model(session, model.UnitOfMeasure,
                                      key='abbreviation')

        for wild_uom in wild_uoms:
            if wild_uom not in known_uoms:
                uom = model.UnitOfMeasure()
                uom.abbreviation = wild_uom
                session.add(uom)

        session.commit()
        session.close()

    def find_wild_uoms(self, session, **kwargs):
        """
        Query some database(s) in order to discover all UOM abbreviations which
        exist "in the wild".

        You are encouraged to override this method as needed.  Note that
        certain POS integration packages may provide some common logic which
        may be used for this.

        :param session: Reference to current Rattail DB session.

        :returns: A list of strings, e.g. ``['OZ', 'LB', ...]``.
        """
        return []


def get_products_handler(config, **kwargs):
    """
    Create and return the configured :class:`ProductsHandler` instance.
    """
    spec = config.get('rattail', 'products.handler')
    if spec:
        factory = load_object(spec)
    else:
        factory = ProductsHandler
    return factory(config, **kwargs)



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Data model for "delete product" batches
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.db.model import Base, BatchMixin, ProductBatchRowMixin


class DeleteProductBatch(BatchMixin, Base):
    """
    Primary data model for "delete product" batches.
    """
    __tablename__ = 'batch_delproduct'
    __batchrow_class__ = 'DeleteProductBatchRow'
    batch_key = 'delproduct'
    model_title = "Delete Product Batch"
    model_title_plural = "Delete Product Batches"

    inactivity_months = sa.Column(sa.Integer(), nullable=True, doc="""
    Number of months of inactivity required for a product to be considered no
    longer active.
    """)


class DeleteProductBatchRow(ProductBatchRowMixin, Base):
    """
    Row of data within a new product batch.
    """
    __tablename__ = 'batch_delproduct_row'
    __batch_class__ = DeleteProductBatch

    STATUS_OK                           = 1
    STATUS_PRODUCT_NOT_FOUND            = 2
    STATUS_DELETE_NOT_ALLOWED           = 3
    STATUS_RECENT_ACTIVITY              = 4
    STATUS_PENDING_CUSTOMER_ORDERS      = 5
    STATUS_HAS_INVENTORY                = 6

    STATUS = {
        STATUS_OK                       : "ok",
        STATUS_PRODUCT_NOT_FOUND        : "product not found",
        STATUS_DELETE_NOT_ALLOWED       : "delete not allowed",
        STATUS_RECENT_ACTIVITY          : "has recent activity",
        STATUS_PENDING_CUSTOMER_ORDERS  : "has pending customer orders",
        STATUS_HAS_INVENTORY            : "has inventory",
    }

    pack_size = sa.Column(sa.Numeric(precision=9, scale=4), nullable=True, doc="""
    Number of *units* which constitute the current *pack* product, if applicable.
    """)

    date_created = sa.Column(sa.Date(), nullable=True, doc="""
    Date on which the product record was originally created, if known.
    """)

    present_in_scale = sa.Column(sa.Boolean(), nullable=True, doc="""
    Flag indicating that this product exists in the scale(s) in addition to the
    primary database(s).
    """)
